"""
This directory holds means to get generic information for specific languages.
"""

# Start ignoring PyUnusedCodeBear
from .Language import Language
from .Language import Languages

from .definitions.Unknown import Unknown
from .definitions.C import C
from .definitions.CPP import CPP
from .definitions.CSharp import CSharp
from .definitions.CSS import CSS
from .definitions.Java import Java
from .definitions.JavaScript import JavaScript
from .definitions.Python import Python
from .definitions.Vala import Vala
from .definitions.html import HTML
# Stop ignoring PyUnusedCodeBear
